import inspect
import re
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Dict

import kedro
from kedro.framework.hooks import hook_impl
from kedro.io import CachedDataSet, DataCatalog
from kedro.pipeline.pipeline import TRANSCODING_SEPARATOR, Pipeline
from kedro.runner import AbstractRunner


def _sub_nonword_chars(data_set_name: str) -> str:
    """Replace non-word characters in data set names since Kedro 0.16.2.

    Args:
        data_set_name: The data set name registered in the data catalog.

    Returns:
        The Kedro-version-dependent name used in `DataCatalog.datasets`.
    """
    if kedro.__version__ >= "0.16.6":
        # https://github.com/quantumblacklabs/kedro/commit/93e4cc3fe7a5ac3d4d7e6a372a2bcb26ce123b85
        return re.sub(r"\W+", "__", data_set_name)
    elif kedro.__version__ >= "0.16.2":
        # https://github.com/quantumblacklabs/kedro/commit/3faa0d454f3584f39285843f1ae28bec18cc3fee
        return re.sub("[^0-9a-zA-Z_]+", "__", data_set_name)
    else:
        return data_set_name


class TeePlugin:
    @hook_impl
    def before_pipeline_run(self, pipeline: Pipeline, catalog: DataCatalog) -> None:
        # Create a copy of the unmodified data catalog for future saves.
        self.physical_catalog = catalog.shallow_copy()

        # Get a reference to the runner object the pipeline is run with.
        runner = next(
            caller[0].f_locals.get("runner")
            for caller in inspect.stack()
            if isinstance(caller[0].f_locals.get("runner"), AbstractRunner)
        )

        # Replace intermediate inputs and outputs with in-memory stores.
        # Transcoding is a case where writing to disk is required, since
        # the act of reading/writing affects the object in the pipeline.
        self.data_set_names = set(catalog.list())
        catalog.add_all(
            {
                name: runner.create_default_data_set(name)
                for name in self.data_set_names - pipeline.inputs() - pipeline.outputs()
                if TRANSCODING_SEPARATOR not in name
            },
            replace=True,
        )

        self.pool = ThreadPoolExecutor()
        self.save_futures = set()

    @hook_impl
    def after_node_run(self, catalog: DataCatalog, outputs: Dict[str, Any]) -> None:
        for name, data in outputs.items():
            if name in self.data_set_names:  # ``DataSet`` is registered
                cleaned_name = _sub_nonword_chars(name)
                physical_dataset = getattr(self.physical_catalog.datasets, cleaned_name)

                # If the intermediate data set was "replaced," persist a
                # copy of the output data with the original save method.
                if getattr(catalog.datasets, cleaned_name) is not physical_dataset:
                    self.save_futures.add(
                        self.pool.submit(self.physical_catalog.save, name, data)
                    )

    def _cleanup(self):
        try:
            for future in as_completed(self.save_futures):
                exception = future.exception()
                if exception:
                    raise exception
        finally:
            self.pool.shutdown(wait=True)

    @hook_impl
    def after_pipeline_run(self) -> None:
        self._cleanup()

    @hook_impl
    def on_pipeline_error(self) -> None:
        self._cleanup()


hooks = TeePlugin()


class CachePlugin:
    @hook_impl
    def before_pipeline_run(self, pipeline: Pipeline, catalog: DataCatalog) -> None:
        # Wrap intermediate inputs and outputs using ``CachedDataSet``s.
        catalog.add_all(
            {
                name: CachedDataSet(getattr(catalog.datasets, _sub_nonword_chars(name)))
                for name in set(catalog.list()) - pipeline.inputs() - pipeline.outputs()
            },
            replace=True,
        )



# Copyright 2020 QuantumBlack Visual Analytics Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
# NONINFRINGEMENT. IN NO EVENT WILL THE LICENSOR OR OTHER CONTRIBUTORS
# BE LIABLE FOR ANY CLAIM, DAMAGES, OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF, OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# The QuantumBlack Visual Analytics Limited ("QuantumBlack") name and logo
# (either separately or in combination, "QuantumBlack Trademarks") are
# trademarks of QuantumBlack. The License does not grant you any right or
# license to the QuantumBlack Trademarks. You may not use the QuantumBlack
# Trademarks or any confusingly similar mark as a trademark for your product,
# or use the QuantumBlack Trademarks in any other manner that might cause
# confusion in the marketplace, including but not limited to in advertising,
# on websites, or on software.
#
# See the License for the specific language governing permissions and
# limitations under the License.

from inspect import getdoc
from pathlib import Path

from setuptools import find_packages, setup

import kedro_accelerator

setup(
    name="kedro-accelerator",
    version=kedro_accelerator.__version__,
    description=getdoc(kedro_accelerator).partition("\n")[0],
    long_description=(Path(__file__) / "../../README.md").resolve().read_text("utf-8"),
    long_description_content_type="text/markdown",
    url="https://github.com/deepyaman/kedro-accelerator",
    author="Deepyaman Datta",
    author_email="deepyaman.datta@utexas.edu",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Kedro",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="kedro",
    packages=find_packages(include=["kedro_accelerator.plugins"]),
    entry_points={
        "kedro.hooks": ["kedro-accelerator = kedro_accelerator.plugins:hooks"]
    },
    python_requires=">=3.6, <3.9",
    install_requires=["kedro>=0.16, <0.18"],
    extras_require={
        "docs": [
            "sphinx>=1.6.3, <2.0",
            "sphinx_rtd_theme==0.4.1",
            "nbsphinx==0.3.4",
            "nbstripout==0.3.3",
            "recommonmark==0.5.0",
            "sphinx-autodoc-typehints==1.6.0",
            "sphinx_copybutton==0.2.5",
            "jupyter_client>=5.1.0, <6.0",
            "tornado>=4.2, <6.0",
            "ipykernel>=4.8.1, <5.0",
        ]
    },
)
